from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from dump_things_service.model import (
    get_model_for_schema,
    get_subclasses,
)
from dump_things_service.resolve_curie import resolve_curie
from dump_things_service.utils import cleaned_json

if TYPE_CHECKING:
    from collections.abc import Iterable

    from pydantic import BaseModel

    from dump_things_service.backends import (
        RecordInfo,
        StorageBackend,
    )
    from dump_things_service.lazy_list import LazyList



submitter_class = 'NCIT_C54269'
submitter_namespace = 'http://purl.obolibrary.org/obo/'


class _ModelStore:
    def __init__(
        self,
        schema: str,
        backend: StorageBackend,
    ):
        self.schema = schema
        self.model = get_model_for_schema(self.schema)[0]
        self.backend = backend

    def store_object(
        self,
        obj: BaseModel,
        submitter: str,
    ) -> Iterable[tuple[str, dict]]:

        if obj.__class__.__name__ == 'Thing':
            msg = f'Cannot store `Thing` instance: {obj}.'
            raise ValueError(msg)

        # Extract inlined records from the object, store individual records
        # and return the list of stored records.
        return [
            (
                obj.__class__.__name__,
                self._store_flat_object(
                    obj=obj,
                    submitter=submitter,
                )
            )
            for obj in self.extract_inlined(obj)
        ]

    def _store_flat_object(
        self,
        obj: BaseModel,
        submitter: str,
    ) -> dict:

        iri = resolve_curie(self.model, obj.pid)
        class_name = obj.__class__.__name__

        json_object = cleaned_json(
            obj.model_dump(exclude_none=True, mode='json'),
            remove_keys=('@type',),
        )

        # Add the submitter id to the record annotations
        self.annotate(json_object, submitter)
        self.backend.add_record(
            iri=iri,
            class_name=class_name,
            json_object=json_object,
        )
        return json_object

    def annotate(
        self,
        json_object: dict,
        submitter: str,
    ) -> None:
        """Add submitter IRI to the record annotations, use CURIE if possible"""
        submitter_iri = self.get_curie(
            submitter_namespace,
            submitter_class,
        )
        if 'annotations' not in json_object:
            json_object['annotations'] = {}
        json_object['annotations'][submitter_iri] = submitter

    def get_curie(
        self,
        name_space: str,
        class_name: str,
    ) -> str:
        prefixes = self.model.linkml_meta.root.get('prefixes')
        if prefixes:
            for prefix_info in prefixes.values():
                if prefix_info['prefix_reference'] == name_space:
                    return f'{prefix_info["prefix_prefix"]}:{class_name}'
        return f'{name_space}{class_name}'

    def extract_inlined(
        self,
        record: BaseModel,
    ) -> list[BaseModel]:

        # The trivial case: no relations
        if not hasattr(record, 'relations') or record.relations is None:
            return [record]

        extracted_sub_records = list(
            chain(
                *[
                    self.extract_inlined(sub_record)
                    for sub_record in record.relations.values()
                    # Do not extract 'empty'-Thing records, those are just
                    # placeholders for already extracted records.
                    if sub_record != self.model.Thing(pid=sub_record.pid)
                ]
            )
        )
        # Simplify the relations in this record. We use "empty" Thing objects
        # as placeholders for extracted records.
        new_record = record.model_copy()
        new_record.relations = {
            sub_record_pid: self.model.Thing(pid=sub_record_pid)
            for sub_record_pid in record.relations
        }
        return [new_record, *extracted_sub_records]

    def get_object_by_pid(
        self,
        pid: str,
    ) -> tuple[str, dict] | tuple[None, None]:
        return self.get_object_by_iri(resolve_curie(self.model, pid))

    def get_object_by_iri(
        self,
        iri: str,
    ) -> tuple[str, dict] | tuple[None, None]:
        record_info = self.backend.get_record_by_iri(iri)
        if record_info:
            return record_info.class_name, record_info.json_object
        return None, None

    def get_objects_of_class(
        self,
        class_name: str,
        *,
        include_subclasses: bool = True,
    ) -> LazyList[RecordInfo]:
        """
        Get all objects of a specific class.

        :param class_name: The name of the class to filter by.
        :return: A lazy list of objects of the specified class and its subclasses.
        """
        if include_subclasses:
            class_names = get_subclasses(self.model, class_name)
        else:
            class_names = [class_name]
        return self.backend.get_records_of_classes(class_names)


_existing_model_stores = {}


def ModelStore(  # noqa: N802
    schema: str,
    backend: StorageBackend,
) -> _ModelStore:
    """
    Create a unique model store for the given schema and backend.

    :param schema: The schema to use for the model store.
    :param backend: The storage backend to use.
    :return: An instance of _ModelStore.
    """
    existing_model_store, _ = _existing_model_stores.get(id(backend), (None, None))
    if not existing_model_store:
        existing_model_store = _ModelStore(schema, backend)
        # We store a pointer to the backend in the value to ensure that the
        # backend object exists while we use its `id` as a key.
        _existing_model_stores[id(backend)] = existing_model_store, backend
    return existing_model_store



from __future__ import annotations

import dataclasses
import enum
import hashlib
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Literal,
)

import yaml
from fastapi import HTTPException
from pydantic import (
    BaseModel,
    ValidationError,
)
from yaml.scanner import ScannerError

from dump_things_service import (
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_404_NOT_FOUND,
)
from dump_things_service.backends.record_dir import RecordDirStore
from dump_things_service.backends.sqlite import SQLiteBackend
from dump_things_service.converter import get_conversion_objects
from dump_things_service.model import get_model_for_schema
from dump_things_service.store.model_store import ModelStore

if TYPE_CHECKING:
    import types


config_file_name = '.dumpthings.yaml'
token_config_file_name = '.token_config.yaml'  # noqa: S105
ignored_files = {'.', '..', config_file_name}


class ConfigError(Exception):
    pass


class MappingMethod(enum.Enum):
    digest_md5 = 'digest-md5'
    digest_md5_p3 = 'digest-md5-p3'
    digest_md5_p3_p3 = 'digest-md5-p3-p3'
    digest_sha1 = 'digest-sha1'
    digest_sha1_p3 = 'digest-sha1-p3'
    digest_sha1_p3_p3 = 'digest-sha1-p3-p3'
    after_last_colon = 'after-last-colon'


class CollectionDirConfig(BaseModel):
    type: Literal['records']
    version: Literal[1]
    schema: str
    format: Literal['yaml']
    idfx: MappingMethod


class TokenPermission(BaseModel):
    curated_read: bool = False
    incoming_read: bool = False
    incoming_write: bool = False


class TokenModes(enum.Enum):
    READ_CURATED = 'READ_CURATED'
    READ_COLLECTION = 'READ_COLLECTION'
    WRITE_COLLECTION = 'WRITE_COLLECTION'
    READ_SUBMISSIONS = 'READ_SUBMISSIONS'
    WRITE_SUBMISSIONS = 'WRITE_SUBMISSIONS'
    SUBMIT = 'SUBMIT'
    SUBMIT_ONLY = 'SUBMIT_ONLY'
    NOTHING = 'NOTHING'


class TokenCollectionConfig(BaseModel):
    mode: TokenModes
    incoming_label: str


class TokenConfig(BaseModel):
    user_id: str
    collections: dict[str, TokenCollectionConfig]


class BackendConfigRecordDir(BaseModel):
    type: Literal['record_dir']


class BackendConfigSQLite(BaseModel):
    type: Literal['sqlite']
    schema: str


class CollectionConfig(BaseModel):
    default_token: str
    curated: Path
    incoming: Path | None = None
    backend: BackendConfigRecordDir | BackendConfigSQLite | None = None


class GlobalConfig(BaseModel):
    type: Literal['collections']
    version: Literal[1]
    collections: dict[str, CollectionConfig]
    tokens: dict[str, TokenConfig]


@dataclasses.dataclass
class InstanceConfig:
    store_path: Path
    collections: dict = dataclasses.field(default_factory=dict)
    stores: dict = dataclasses.field(default_factory=dict)
    curated_stores: dict = dataclasses.field(default_factory=dict)
    incoming: dict = dataclasses.field(default_factory=dict)
    zones: dict = dataclasses.field(default_factory=dict)
    model_info: dict = dataclasses.field(default_factory=dict)
    token_stores: dict = dataclasses.field(default_factory=dict)
    schemas: dict = dataclasses.field(default_factory=dict)
    conversion_objects: dict = dataclasses.field(default_factory=dict)
    backend: dict = dataclasses.field(default_factory=dict)


mode_mapping = {
    TokenModes.READ_CURATED: TokenPermission(curated_read=True),
    TokenModes.READ_COLLECTION: TokenPermission(curated_read=True, incoming_read=True),
    TokenModes.WRITE_COLLECTION: TokenPermission(
        curated_read=True, incoming_read=True, incoming_write=True
    ),
    TokenModes.READ_SUBMISSIONS: TokenPermission(incoming_read=True),
    TokenModes.WRITE_SUBMISSIONS: TokenPermission(
        incoming_read=True, incoming_write=True
    ),
    TokenModes.SUBMIT: TokenPermission(curated_read=True, incoming_write=True),
    TokenModes.SUBMIT_ONLY: TokenPermission(incoming_write=True),
    TokenModes.NOTHING: TokenPermission(),
}


def get_hex_digest(hasher: Callable, data: str) -> str:
    hash_context = hasher(data.encode())
    return hash_context.hexdigest()


def mapping_digest_p3(
    hasher: Callable,
    pid: str,
    suffix: str,
) -> Path:
    hex_digest = get_hex_digest(hasher, pid)
    return Path(hex_digest[:3]) / (hex_digest[3:] + '.' + suffix)


def mapping_digest_p3_p3(
        hasher: Callable,
        pid: str,
        suffix: str,
) -> Path:
    hex_digest = get_hex_digest(hasher, pid)
    return Path(hex_digest[:3]) / hex_digest[3:6] / (hex_digest[6:] + '.' + suffix)


def mapping_digest(hasher: Callable, pid: str, suffix: str) -> Path:
    hex_digest = get_hex_digest(hasher, pid)
    return Path(hex_digest + '.' + suffix)


def mapping_after_last_colon(pid: str, suffix: str) -> Path:
    plain_result = pid.split(':')[-1]
    # Escape any colons and slashes in the pid
    escaped_result = (
        plain_result.replace('_', '__').replace('/', '_s').replace('.', '_d')
    )
    return Path(escaped_result + '.' + suffix)


mapping_functions = {
    MappingMethod.digest_md5: partial(mapping_digest, hashlib.md5),
    MappingMethod.digest_md5_p3: partial(mapping_digest_p3, hashlib.md5),
    MappingMethod.digest_md5_p3_p3: partial(mapping_digest_p3_p3, hashlib.md5),
    MappingMethod.digest_sha1: partial(mapping_digest, hashlib.sha1),
    MappingMethod.digest_sha1_p3: partial(mapping_digest_p3, hashlib.sha1),
    MappingMethod.digest_sha1_p3_p3: partial(mapping_digest_p3_p3, hashlib.sha1),
    MappingMethod.after_last_colon: mapping_after_last_colon,
}


def get_mapping_function_by_name(mapping_function_name: str) -> Callable:
    return mapping_functions[MappingMethod(mapping_function_name)]


def get_mapping_function(collection_config: CollectionDirConfig):
    return mapping_functions[collection_config.idfx]


def get_permissions(mode: TokenModes) -> TokenPermission:
    return mode_mapping[mode]


class Config:
    @staticmethod
    def get_config_from_file(path: Path) -> GlobalConfig:
        try:
            return GlobalConfig(**yaml.load(path.read_text(), Loader=yaml.SafeLoader))
        except ScannerError as e:
            msg = f'YAML-error while reading config file {path}: {e}'
            raise ConfigError(msg) from e
        except TypeError:
            msg = f'Error in yaml file {path}: content is not a mapping'
            raise ConfigError(msg) from None
        except ValidationError as e:
            msg = f'Pydantic-error reading config file {path}: {e}'
            raise ConfigError(msg) from e

    @staticmethod
    def get_config(path: Path, file_name=config_file_name) -> GlobalConfig:
        return Config.get_config_from_file(path / file_name)

    @staticmethod
    def get_collection_dir_config(
        path: Path,
        file_name: str = config_file_name,
    ) -> CollectionDirConfig:
        config_path = path / file_name
        try:
            return CollectionDirConfig(
                **yaml.load(config_path.read_text(), Loader=yaml.SafeLoader)
            )
        except ScannerError as e:
            msg = f'YAML-error while reading config file {config_path}: {e}'
            raise ConfigError(msg) from e
        except ValidationError as e:
            msg = f'Pydantic-error reading config file {config_path}: {e}'
            raise ConfigError(msg) from e


def process_config(
    store_path: Path,
    config_file: Path,
    order_by: list[str],
    globals_dict: dict[str, Any],
    *,
    create_models: bool = True,
) -> InstanceConfig:

    config_object = Config.get_config_from_file(config_file)
    return process_config_object(
        store_path=store_path,
        config_object=config_object,
        order_by=order_by,
        globals_dict=globals_dict,
        create_models=create_models,
    )


def process_config_object(
    store_path: Path,
    config_object: GlobalConfig,
    order_by: list[str],
    globals_dict: dict[str, Any],
    *,
    create_models: bool = True,
):

    instance_config = InstanceConfig(store_path=store_path)
    instance_config.collections = config_object.collections

    # Create a `ModelStore` (with currently fixed backend `RecordDirStore`) for
    # the `curated`-dir in each collection.
    for collection_name, collection_info in config_object.collections.items():
        backend = collection_info.backend or BackendConfigRecordDir(type='record_dir')
        instance_config.backend[collection_name] = backend
        if backend.type == 'record_dir':
            # Get the config from the curated directory
            collection_config = Config.get_collection_dir_config(store_path / collection_info.curated)
            schema = collection_config.schema
        elif backend.type == 'sqlite':
            schema = backend.schema
        else:
            msg = f'Unsupported backend `{collection_info.backend}` for collection `{collection_name}`.'
            raise ConfigError(msg)

        # Generate the collection model
        if create_models:
            model, classes, model_var_name = get_model_for_schema(schema)
            instance_config.model_info[collection_name] = model, classes, model_var_name
            globals_dict[model_var_name] = model

        if backend.type == 'record_dir':
            collection_config = Config.get_collection_dir_config(store_path / collection_info.curated)
            curated_store_backend = RecordDirStore(
                root=store_path / collection_info.curated,
                schema=schema,
                pid_mapping_function=get_mapping_function(collection_config),
                suffix=collection_config.format,
                order_by=order_by,
            )
        elif backend.type == 'sqlite':
            curated_store_backend = SQLiteBackend(
                db_path=store_path / collection_info.curated / 'records.db',
            )
        else:
            msg = f'Unsupported backend `{collection_info.backend}` for collection `{collection_name}`.'
            raise ConfigError(msg)

        curated_store = ModelStore(
            schema=schema,
            backend=curated_store_backend,
        )

        instance_config.curated_stores[collection_name] = curated_store

        if collection_info.incoming:
            instance_config.incoming[collection_name] = collection_info.incoming

        instance_config.schemas[collection_name] = schema
        if create_models and schema not in instance_config.conversion_objects:
            instance_config.conversion_objects[schema] = get_conversion_objects(schema)

    # Create a `ModelStore` for each token dir and fetch the permissions
    for token_name, token_info in config_object.tokens.items():
        entry = {'user_id': token_info.user_id, 'collections': {}}
        instance_config.token_stores[token_name] = entry
        for collection_name, token_collection_info in token_info.collections.items():
            backend = instance_config.backend[collection_name]
            entry['collections'][collection_name] = {}

            # A token might be a pure curated read token, i.e., have the mode
            # `READ_COLLECTION`. In this case, there might be no incoming store.
            if (
                    collection_name in instance_config.incoming and
                    token_collection_info.mode not in (
                        TokenModes.READ_CURATED,
                        TokenModes.NOTHING,
                    )
            ):
                # Check that the incoming label is set for a token that has
                # access rights to incoming records.
                if not token_collection_info.incoming_label:
                    msg = f'Token `{token_name}` with mode {token_collection_info.mode} must not have an empty `incoming_label`'
                    raise ConfigError(msg)

                if any(c in token_collection_info.incoming_label for c in ('\\', '/')):
                    msg = f'Incoming label for token `{token_name}` must not contain slashes or backslashes: `{token_collection_info.incoming_label}`'
                    raise ConfigError(msg)

                if collection_name not in instance_config.zones:
                    instance_config.zones[collection_name] = {}
                instance_config.zones[collection_name][token_name] = token_collection_info.incoming_label
                # Ensure that the store directory exists
                store_dir = (
                        store_path
                        / instance_config.incoming[collection_name]
                        / token_collection_info.incoming_label
                )
                store_dir.mkdir(parents=True, exist_ok=True)
                if backend.type == 'record_dir':
                    mapping_function = instance_config.curated_stores[collection_name].backend.pid_mapping_function
                    token_store_backend = RecordDirStore(
                        root=store_dir,
                        schema=instance_config.schemas[collection_name],
                        pid_mapping_function=mapping_function,
                        suffix=instance_config.curated_stores[collection_name].backend.suffix,
                        order_by=order_by,
                    )
                elif backend.type == 'sqlite':
                    token_store_backend = SQLiteBackend(
                        db_path=store_dir / 'records.db',
                    )
                else:
                    msg = f'Unsupported backend `{collection_info.backend.type}` for collection `{collection_name}`.'
                    raise ConfigError(msg)
                token_store = ModelStore(
                    schema=instance_config.schemas[collection_name],
                    backend=token_store_backend,
                )
                entry['collections'][collection_name]['store'] = token_store

            entry['collections'][collection_name]['permissions'] = get_permissions(
                token_collection_info.mode
            )
    return instance_config


def get_token_store(
        instance_config: InstanceConfig,
        collection_name: str, token: str
) -> tuple[ModelStore, TokenPermission] | tuple[None, None]:
    if collection_name not in instance_config.curated_stores:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f'No such collection: "{collection_name}".',
        )
    if token not in instance_config.token_stores:
        raise HTTPException(status_code=HTTP_401_UNAUTHORIZED, detail='Invalid token.')

    token_store = None
    token_collection_info = instance_config.token_stores[token]['collections'][collection_name]
    permissions = token_collection_info['permissions']
    if permissions.incoming_write or permissions.incoming_read:
        token_store = token_collection_info.get('store')
        if not token_store:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=f'Configuration does not define an incoming store for token "{token}" in collection "{collection_name}".',
            )
    return token_store, permissions


def get_default_token_name(
        instance_config: InstanceConfig,
        collection: str
) -> str:
    if collection not in instance_config.collections:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail=f'No such collection: {collection}'
        )
    return instance_config.collections[collection].default_token


def join_default_token_permissions(
        instance_config: InstanceConfig,
        permissions: TokenPermission,
        collection: str,
) -> TokenPermission:
    default_token_name = instance_config.collections[collection].default_token
    default_token_permissions = instance_config.token_stores[default_token_name]['collections'][collection]['permissions']
    result = TokenPermission()
    result.curated_read = permissions.curated_read | default_token_permissions.curated_read
    result.incoming_read = permissions.incoming_read | default_token_permissions.incoming_read
    result.incoming_write = permissions.incoming_write | default_token_permissions.incoming_write
    return result


def get_zone(
        instance_config: InstanceConfig,
        collection: str,
        token: str,
) -> str | None:
    """Get the zone for the given collection and token."""
    if collection not in instance_config.zones:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail=f'No incoming zone defined for collection: {collection}'
        )
    if token not in instance_config.zones[collection]:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail=f'No incoming zone defined for collection: {collection}'
        )
    return instance_config.zones[collection][token]


def get_conversion_objects_for_collection(
    instance_config: InstanceConfig,
    collection_name: str,
) -> dict:
    """Get the conversion objects for the given collection."""
    if collection_name not in instance_config.schemas:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail=f'No such collection: {collection_name}'
        )
    return instance_config.conversion_objects[instance_config.schemas[collection_name]]


def get_model_info_for_collection(
    instance_config: InstanceConfig,
    collection_name: str,
) -> tuple[types.ModuleType, dict[str, Any], str]:
    if collection_name not in instance_config.model_info:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail=f'No such collection: {collection_name}'
        )
    return instance_config.model_info[collection_name]



""" Monkeypatch for linkml_runtime.utils.yamlutil.YAMLRoot._normalize_inlined

Corresponds to the patch `patches/linkml_runtime_utils_yamlutils.diff` in the
`datalad-concepts` repository, i.e.,
<https://github.com/psychoinformatics-de/datalad-concepts.git>
"""

from copy import copy
from importlib import import_module
from typing import (
    Any,
    Optional,
    Union,
)

from jsonasobj2 import (
    JsonObj,
    JsonObjTypes,
    as_dict,
)
from linkml_runtime.utils.formatutils import items
from linkml_runtime.utils.yamlutils import (
    TypedNode,
    YAMLRoot,
)

YAMLObjTypes = Union[JsonObjTypes, "YAMLRoot"]

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    pass



def _normalize_inlined(self, slot_name: str, slot_type: type, key_name: str, keyed: bool, is_list: bool) \
        -> None:
    """
     __post_init__ function for a list of inlined keyed or identified classes.

    The input to this is either a list or dictionary of dictionaries.  In the list case, every key entry
    in the list must be unique.  In the dictionary case, the key may or may not be repeated in the dictionary
    element. The internal storage structure is a dictionary of dictionaries.
    @param slot_name: Name of the slot being normalized
    @param slot_type: Slot range type
    @param key_name: Name of the key or identifier in the range
    @param keyed: True means each identifier must be unique
    @param is_list: True means inlined as list
    """
    raw_slot: Union[list, dict, JsonObj] = self[slot_name]
    if raw_slot is None:
        raw_slot = []
    elif not isinstance(raw_slot, (dict, list, JsonObj)):
        raw_slot = [raw_slot]
    cooked_slot = list() if is_list else dict()
    cooked_keys = set()

    def order_up(key: Any, cooked_entry: YAMLRoot) -> None:
        """ A cooked entry is ready to be added to the return slot """
        if cooked_entry[key_name] != key:
            raise ValueError(
                f"Slot: {loc(slot_name)} - attribute {loc(key_name)} " \
                f"value ({loc(cooked_entry[key_name])}) does not match key ({loc(key)})")
        if keyed and key in cooked_keys:
            raise ValueError(f"{loc(key)}: duplicate key")
        cooked_keys.add(key)
        if is_list:
            cooked_slot.append(cooked_entry)
        else:
            cooked_slot[key] = cooked_entry

    def loc(s):
        loc_str = TypedNode.yaml_loc(s) if isinstance(s, TypedNode) else ''
        if loc_str == ': ':
            loc_str = ''
        return loc_str + str(s)

    def form_1(entries: dict[Any, Optional[Union[dict, JsonObj]]]) -> None:
        """ A dictionary of key:dict entries where key is the identifier and dict is an instance of slot_type """
        for key, raw_obj in items(entries):
            if raw_obj is None:
                raw_obj = {}
            if key_name not in raw_obj:
                raw_obj = copy(raw_obj)
                raw_obj[key_name] = key
            if not issubclass(type(raw_obj), slot_type):
                order_up(key, slot_type(**as_dict(raw_obj)))
            else:
                order_up(key, raw_obj)

    # TODO: Make an external function extract a root JSON list
    if isinstance(raw_slot, JsonObj):
        raw_slot = raw_slot._hide_list()

    if isinstance(raw_slot, list):
        # We have a list of entries
        for list_entry in raw_slot:
            if isinstance(list_entry, slot_type):
                order_up(list_entry[key_name], list_entry)
            elif isinstance(list_entry, (dict, JsonObj)):
                # list_entry is either a key:dict, key_name:value or **kwargs
                if len(list_entry) == 1:
                    # key:dict or key_name:key
                    for lek, lev in items(list_entry):
                        if lek == key_name and not isinstance(lev, (list, dict, JsonObj)):
                            # key_name:value
                            order_up(list_entry[lek], slot_type(**list_entry))
                            break   # Not strictly necessary, but
                        elif not isinstance(lev, (list, dict, JsonObj)):
                            # key: value --> slot_type(key, value)
                            order_up(lek, slot_type(lek, lev))
                        else:
                            form_1(list_entry)
                else:
                    # **kwargs
                    cooked_obj = slot_type(**as_dict(list_entry))
                    order_up(cooked_obj[key_name], cooked_obj)
            elif isinstance(list_entry, list):
                # *args
                cooked_obj = slot_type(*list_entry)
                order_up(cooked_obj[key_name], cooked_obj)
            else:
                # lone key [key1: , key2: ... }
                order_up(list_entry, slot_type(**{key_name: list_entry}))
    elif key_name in raw_slot and raw_slot[key_name] is not None \
                and not isinstance(raw_slot[key_name], (list, dict, JsonObj)):
        # Vanilla dictionary - {key: v11, s12: v12, ...}
        order_up(raw_slot[key_name], slot_type(**as_dict(raw_slot)))
    else:
        # We have either {key1: {obj1}, key2: {obj2}...} or {key1:, key2:, ...}
        for k, v in items(raw_slot):
            if v is None:
                v = dict()
            if isinstance(v, slot_type):
                order_up(k, v)
            elif isinstance(v, (dict, JsonObj)):
                form_1({k: v})
            elif not isinstance(v, list):
                order_up(k, slot_type(*[k, v]))
            else:
                raise ValueError(f"Unrecognized entry: {loc(k)}: {v!s}")
    self[slot_name] = cooked_slot


mod = import_module('linkml_runtime.utils.yamlutils')
mod.YAMLRoot._normalize_inlined = _normalize_inlined



"""
Backend that stores records in a directory structure

The disk-layout is described in <https://concepts.datalad.org/dump-things/>.

The backend has a special handling for `schema_type` attributes. Because the
class of the record is encoded in the record-path, the `schema_type` attribute
in the top level dictionary is redundant. Therefore, it is removed in
stored records, i.e., it is not stored in the stored YAML files.

When a record is read from disk, the `schema_type` attribute is added in all
cases (because we don't keep track of whether the initial record had a
`schema_type`-attribute or not). So every record that is read from this backend
will contain a `schema_type` attribute.
"""
from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Callable,
)

import yaml

from dump_things_service import config_file_name
from dump_things_service.backends import (
    BackendResultList,
    RecordInfo,
    ResultListInfo,
    StorageBackend,
    create_sort_key,
)
from dump_things_service.model import (
    get_model_for_schema,
    get_schema_model_for_schema,
)
from dump_things_service.resolve_curie import resolve_curie

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path
    from types import ModuleType


__all__ = [
    'RecordDirStore',
]

ignored_files = {'.', '..', config_file_name}

lgr = logging.getLogger('dump_things_service')


class RecordDirResultList(BackendResultList):
    """
    The specific result list for record directory backends.
    """
    def __init__(self, schema_model: ModuleType):
        """
        Initialize the record list.

        :param schema_model: The python schema module to use for schema type resolution.
        """
        super().__init__()
        self.schema_model = schema_model

    def generate_result(
        self,
        _: int,
        iri: str,
        class_name: str,
        sort_key: str,
        path: Path,
    ) -> RecordInfo:
        """
        Generate a JSON representation of the record at index `index`.

        :param _: The index of the record.
        :param iri: The IRI of the record.
        :param class_name: The class name of the record.
        :param sort_key: The sort key for the record.
        :param path: The path where the record is stored
        :return: A RecordInfo object.
        """
        with path.open('r') as f:
            json_object = yaml.load(f, Loader=yaml.SafeLoader)
            json_object['schema_type'] = _get_schema_type(class_name, self.schema_model)
            return RecordInfo(
                iri=iri,
                class_name=class_name,
                json_object=json_object,
                sort_key=sort_key,
            )


class _RecordDirStore(StorageBackend):
    """Store records in a directory structure"""

    def __init__(
        self,
        root: Path,
        # `schema` is required for two purposes:
        #  1. CURIE prefix resolution during index creation
        #  2. `schema_type` determination for records that are read from disk
        schema: str,
        pid_mapping_function: Callable,
        suffix: str,
        order_by: Iterable[str] | None = None,
    ):
        super().__init__(order_by=order_by)
        if not root.is_absolute():
            msg = f'Store root is not absolute: {root}'
            raise ValueError(msg)
        self.root = root
        self.schema = schema
        self.model = get_model_for_schema(self.schema)[0]
        self.schema_model = get_schema_model_for_schema(self.schema)
        self.pid_mapping_function = pid_mapping_function
        self.suffix = suffix
        self.index = {}
        self._build_index()

    def _build_index(self):
        lgr.info('Building IRI index for records in %s', self.root)
        for path in self.root.rglob(f'*.{self.suffix}'):
            if path.is_file() and path.name not in ignored_files:
                try:
                    # Catch YAML structure errors
                    record = yaml.load(path.read_text(), Loader=yaml.SafeLoader)
                except Exception as e:  # noqa: BLE001
                    lgr.error('Error: reading YAML record from %s: %s', path, e)
                    continue

                try:
                    # Catch YAML payload errors
                    pid = record['pid']
                except (TypeError, KeyError):
                    lgr.error(
                        'Error: record at %s does not contain a mapping with `pid`',
                        path,
                    )
                    continue

                iri = resolve_curie(self.model, pid)
                sort_string = create_sort_key(record, self.order_by)

                # On startup, log PID collision errors and continue building the index
                try:
                    self._add_iri_to_index(
                        iri, self._get_class_name(path), pid, path, sort_string
                    )
                except ValueError as e:
                    lgr.error('Error during index creation: %s', e)
        lgr.info('Index built with %d IRIs', len(self.index))

    def _add_iri_to_index(
        self,
        iri: str,
        new_class: str,
        pid: str,
        path: Path,
        sort_string: str,
    ):
        # If the IRI is already in the index, the reasons may be:
        #
        # 1. The existing record is updated. In this case the path should
        #    be the same as the one already in the index (which means the classes
        #    are the same and the PIDs are the same). No need to replace the path
        #    since they are identical anyway.
        # 2. The existing record is a different class (not `Thing`) and probably
        #    a different PID. That indicates that two different records have the
        #    same IRI. This is an error condition, and we raise an exception
        existing_entry = self.index.get(iri)
        if existing_entry:
            existing_class, existing_pid, existing_path, existing_sort_string = existing_entry
            # Case 1: existing record is updated
            if path == existing_path:
                self.index[iri] = existing_class, pid, path, sort_string
                return
            # Case 2:
            msg = f'Duplicated IRI ({iri}): already indexed {existing_class}-instance at {existing_path} has the same IRI as new {new_class}-instance at {path}.'
            raise ValueError(msg)
        self.index[iri] = new_class, pid, path, sort_string

    def _get_class_name(self, path: Path) -> str:
        """Get the class name from the path."""
        rel_path = path.absolute().relative_to(self.root)
        return rel_path.parts[0]

    def rebuild_index(self):
        self.index = {}
        self._build_index()

    def add_record(
        self,
        iri: str,
        class_name: str,
        json_object: dict,
    ):
        pid = json_object['pid']

        # Generate the class directory, apply the mapping function to the record
        # pid to get the final storage path.
        record_root = self.root / class_name
        record_root.mkdir(exist_ok=True)
        storage_path = record_root / self.pid_mapping_function(pid=pid, suffix='yaml')

        # Ensure that the storage path is within the record root
        try:
            storage_path.relative_to(record_root)
        except ValueError as e:
            msg = (
                f'Invalid pid ({pid}) for mapping function: {self.pid_mapping_function}'
            )
            raise ValueError(msg) from e

        # Ensure all intermediate directories exist and save the YAML document
        storage_path.parent.mkdir(parents=True, exist_ok=True)

        # Remove the top level `schema_type` from the JSON object because we
        # don't want to store it in the YAML file. We add `schema_type` after
        # reading the record from disk. The value of `schema_type` is determined
        # by the class name of the record, which is stored in the path.
        if 'schema_type' in json_object:
            del json_object['schema_type']

        # Convert the record object into a YAML object
        data = yaml.dump(
            data=json_object,
            sort_keys=False,
            allow_unicode=True,
            default_flow_style=False,
        )
        storage_path.write_text(data, encoding='utf-8')

        # Add the IRI to the index
        sort_string = create_sort_key(json_object, self.order_by)
        self._add_iri_to_index(iri, class_name, pid, storage_path, sort_string)

    def get_record_by_iri(
        self,
        iri: str,
    ) -> RecordInfo | None:
        index_entry = self.index.get(iri)
        if index_entry is None:
            return None

        class_name, pid, path, sort_key = index_entry
        json_object = yaml.load(path.read_text(), Loader=yaml.SafeLoader)
        json_object['schema_type'] = _get_schema_type(class_name, self.schema_model)
        return RecordInfo(
            iri=iri,
            class_name=class_name,
            json_object=json_object,
            sort_key=sort_key,
        )

    def get_records_of_classes(
        self,
        class_names: list[str]
    ) -> RecordDirResultList:
        return RecordDirResultList(self.schema_model).add_info(
            sorted(
                (
                    ResultListInfo(
                        iri=iri,
                        class_name=class_name,
                        sort_key=sort_key,
                        private=path,
                    )
                    for iri, (class_name, pid, path, sort_key) in self.index.items()
                    if class_name in class_names
                ),
                key=lambda result_list_info: result_list_info.sort_key,
            )
        )


# Ensure that there is only one store per root directory.
_existing_stores = {}

def RecordDirStore(  # noqa: N802
        root: Path,
        schema: str,
        pid_mapping_function: Callable,
        suffix: str,
        order_by: Iterable[str] | None = None,
) -> _RecordDirStore:
    """Get a record directory store for the given root directory."""
    existing_store = _existing_stores.get(root)
    if not existing_store:
        existing_store = _RecordDirStore(
            root=root,
            schema=schema,
            pid_mapping_function=pid_mapping_function,
            suffix=suffix,
            order_by=order_by,
        )
        _existing_stores[root] = existing_store

    if existing_store.schema != schema:
        msg = f'Store at {root} already exists with different schema.'
        raise ValueError(msg)

    if existing_store.pid_mapping_function != pid_mapping_function:
        msg = f'Store at {root} already exists with different PID mapping function.'
        raise ValueError(msg)

    if existing_store.suffix != suffix:
        msg = f'Store at {root} already exists with different format.'
        raise ValueError(msg)

    if existing_store.order_by != (order_by or ['pid']):
        msg = f'Store at {root} already exists with different order specification.'
        raise ValueError(msg)

    return existing_store


def _get_schema_type(
    class_name: str,
    schema_module: ModuleType,
) -> str:
    return getattr(schema_module, class_name).class_class_curie



from enum import Enum
from typing import (
    Any,
    Union,
)

from starlette.status import (
    HTTP_200_OK,
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_413_REQUEST_ENTITY_TOO_LARGE,
    HTTP_500_INTERNAL_SERVER_ERROR,
)

__all__ = [
    'Format',
    'HTTP_200_OK',
    'HTTP_400_BAD_REQUEST',
    'HTTP_401_UNAUTHORIZED',
    'HTTP_403_FORBIDDEN',
    'HTTP_404_NOT_FOUND',
    'HTTP_413_REQUEST_ENTITY_TOO_LARGE',
    'HTTP_500_INTERNAL_SERVER_ERROR',
    'JSON',
    'YAML',
    'config_file_name',
]


class Format(str, Enum):
    json = 'json'
    ttl = 'ttl'


JSON = Union[dict[str, Any], list[Any], str, int, float, None]
YAML = JSON

config_file_name = '.dumpthings.yaml'



"""
SQLAlchemy backend

Disk usage with sqlite-driver:

JSON string with about 122 characters per record:

10.000 records: 3 MB
100.000 records: 30 MB
1.000.000 records: 310 MB

roughly 300 bytes per record


JSON string with about 244 characters per record:

10.000 records: 4 MB
100.000 records: 43 MB
1.000.000 records: 431 MB

roughly 400 bytes per record

Presumably, 180 bytes + JSON string size per record.

"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
)

from sqlalchemy import (
    JSON,
    String,
    create_engine,
    select,
)
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    Session,
    load_only,
    mapped_column,
)

from dump_things_service.backends import (
    BackendResultList,
    RecordInfo,
    ResultListInfo,
    StorageBackend,
    create_sort_key,
)

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path


class Base(DeclarativeBase):
    pass


class Thing(Base):
    __tablename__ = 'thing'

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    iri: Mapped[str] = mapped_column(nullable=False, unique=True, index=True)
    class_name: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    object: Mapped[dict] = mapped_column(JSON, nullable=False)
    sort_key: Mapped[str] = mapped_column(nullable=False)


class SQLResultList(BackendResultList):
    def __init__(
        self,
        engine: Any,
    ):
        super().__init__()
        self.engine = engine

    def generate_result(
            self,
            _: int,
            iri: str,
            class_name: str,
            sort_key: str,
            db_id: int,
    ) -> RecordInfo:
        """
        Generate a JSON representation of the record at index `index`.

        :param _: The index of the record.
        :param iri: The IRI of the record.
        :param class_name: The class name of the record.
        :param sort_key: The sort key for the record.
        :param db_id: The id of the record in the database
        :return: A RecordInfo object.
        """
        with Session(self.engine) as session, session.begin():
            thing = session.get(Thing, db_id)
            return RecordInfo(
                iri=iri, class_name=class_name, json_object=thing.object, sort_key=sort_key
            )


class _SQLiteBackend(StorageBackend):
    def __init__(
        self,
        db_path: Path,
        *,
        order_by: Iterable[str] | None = None,
        echo: bool = False
    ) -> None:
        super().__init__(order_by=order_by)
        self.engine = create_engine('sqlite:///' + str(db_path), echo=echo)
        Base.metadata.create_all(self.engine)

    def add_record(
        self,
        iri: str,
        class_name: str,
        json_object: dict,
    ):
        with Session(self.engine) as session, session.begin():
            self._add_record_with_session(
                session=session,
                iri=iri,
                class_name=class_name,
                json_object=json_object,
            )

    def add_records_bulk(
        self,
        record_infos: Iterable[RecordInfo],
    ):
        with Session(self.engine) as session, session.begin():
            for record_info in record_infos:
                self._add_record_with_session(
                    session=session,
                    iri=record_info.iri,
                    class_name=record_info.class_name,
                    json_object=record_info.json_object,
                )

    def _add_record_with_session(
            self,
            session: Session,
            iri: str,
            class_name: str,
            json_object: dict,
    ):
        sort_key = create_sort_key(json_object, self.order_by)
        existing_record = session.query(Thing).filter_by(iri=iri).first()
        if existing_record:
            existing_record.class_name = class_name
            existing_record.object = json_object
            existing_record.sort_key=sort_key
        else:
            session.add(
                Thing(
                    iri=iri,
                    class_name=class_name,
                    object=json_object,
                    sort_key=sort_key,
                )
            )

    def get_record_by_iri(
        self,
        iri: str,
    ) -> RecordInfo | None:
        with Session(self.engine) as session, session.begin():
            statement = select(Thing).filter_by(iri=iri)
            thing = session.scalar(statement)
            if thing:
                return RecordInfo(
                    iri=thing.iri,
                    class_name=thing.class_name,
                    json_object=thing.object,
                    sort_key=thing.sort_key,
                )
        return None

    def get_records_of_classes(
        self,
        class_names: Iterable[str],
    ) -> SQLResultList:
        statement = select(Thing).where(Thing.class_name.in_(class_names))
        for attribute in self.order_by:
            statement = statement.order_by(Thing.object[attribute]).options(
                load_only(Thing.id)
            )
        with Session(self.engine) as session, session.begin():
            return SQLResultList(self.engine).add_info(
                ResultListInfo(
                    iri=thing.iri,
                    class_name=thing.class_name,
                    sort_key=thing.sort_key,
                    private=thing.id,
                )
                for thing in session.scalars(statement).all()
            )


# Ensure that there is only one SQL-backend per database file.
_existing_sqlite_backends = {}


def SQLiteBackend(  # noqa: N802
        db_path: Path,
        *,
        order_by: Iterable[str] | None = None,
        echo: bool = False
) -> _SQLiteBackend:
    existing_backend = _existing_sqlite_backends.get(db_path)
    if not existing_backend:
        existing_backend = _SQLiteBackend(
            db_path=db_path,
            order_by=order_by,
            echo=echo,
        )
        _existing_sqlite_backends[db_path] = existing_backend

    if existing_backend.order_by != (order_by or ['pid']):
        msg = f'Store at {db_path} already exists with different order specification.'
        raise ValueError(msg)

    return existing_backend



from __future__ import annotations

import re
from json import loads as json_loads
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
)

from linkml.utils.datautils import (
    get_dumper,
    get_loader,
)
from linkml_runtime import SchemaView
from rdflib.term import (
    URIRef,
    bind,
)

from dump_things_service import Format
from dump_things_service.lazy_list import LazyList
from dump_things_service.model import (
    get_model_for_schema,
    get_schema_model_for_schema,
)
from dump_things_service.utils import cleaned_json

if TYPE_CHECKING:
    from types import ModuleType

    from pydantic import BaseModel

    from dump_things_service.backends import RecordInfo


_cached_conversion_objects = {}


# Enable rdflib to parse date time literals with type ref:
# `https://www.w3.org/TR/NOTE-datetime`. We add a regex-based validation and
# return the validated datetime string verbatim.

_datetime_regex = re.compile(r'^([-+]\d+)|(\d{4})|(\d{4}-[01]\d)|(\d{4}-[01]\d-[0-3]\d)|(\d{4}-[01]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d\.\d+([+-][0-2]\d:[0-5]\d|Z))|(\d{4}-[01]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d([+-][0-2]\d:[0-5]\d|Z))|(\d{4}-[01]\d-[0-3]\dT[0-2]\d:[0-5]\d([+-][0-2]\d:[0-5]\d|Z))$')


def _validate_datetime(value: str) -> str:
    match = _datetime_regex.match(value)
    if not match:
        msg = 'Invalid datetime format: {value}'
        raise ValueError(msg)
    return value

bind(
    datatype=URIRef('https://www.w3.org/TR/NOTE-datetime'),
    constructor=_validate_datetime,
    pythontype=str,
)


def get_conversion_objects(schema: str):
    if schema not in _cached_conversion_objects:
        _cached_conversion_objects[schema] = {
            'schema_module': get_schema_model_for_schema(schema),
            'schema_view': SchemaView(schema),
        }
    return _cached_conversion_objects[schema]


class FormatConverter:
    def __init__(
        self,
        schema: str,
        input_format: Format,
        output_format: Format,
    ):
        self.converter = self._check_formats(input_format, output_format)
        self.model = get_model_for_schema(schema)[0]
        self.conversion_objects = get_conversion_objects(schema)

    def _check_formats(
        self,
        input_format: Format,
        output_format: Format,
    ) -> Callable:
        if input_format == output_format:
            return lambda data, _: data
        if input_format == Format.ttl:
            return self._convert_ttl_to_json
        return self._convert_json_to_ttl

    def convert(
        self,
        data: str | dict,
        target_class: str
    ) -> str | dict:
        return self.converter(data, target_class)

    def _convert_json_to_ttl(
        self,
        data: dict,
        target_class: str,
    ) -> str:
        pydantic_object = getattr(self.model, target_class)(**data)
        return self._convert_pydantic_to_ttl(pydantic_object=pydantic_object)

    def _convert_pydantic_to_ttl(
        self,
        pydantic_object: BaseModel,
    ):
        return _convert_format(
            target_class=pydantic_object.__class__.__name__,
            data=pydantic_object.model_dump(mode='json', exclude_none=True),
            input_format=Format.json,
            output_format=Format.ttl,
            **self.conversion_objects,
        )

    def _convert_ttl_to_json(
        self,
        data: str,
        target_class: str,
    ) -> dict:
        json_string = _convert_format(
            target_class=target_class,
            data=data,
            input_format=Format.ttl,
            output_format=Format.json,
            **self.conversion_objects
        )
        return cleaned_json(json_loads(json_string))


class ConvertingList(LazyList):
    """
    A lazy list that converts records stored in an "input" lazy list. The
    input lazy list must return `RecordInfo`-objects.
    """
    def __init__(self,
        input_list: LazyList,
        schema: str,
        input_format: Format,
        output_format: Format,
    ):
        super().__init__()
        self.input_list = input_list
        # We reuse `list_info` from the input list to save time and memory.
        self.list_info = input_list.list_info
        self.converter = FormatConverter(schema, input_format, output_format)

    def generate_element(self, index: int, _: Any) -> Any:
        record_info: RecordInfo = self.input_list[index]
        record_info.json_object = self.converter.convert(
            data=record_info.json_object,
            target_class=record_info.class_name,
        )
        return record_info.json_object


def _convert_format(
    target_class: str,
    data: dict | str,
    input_format: Format,
    output_format: Format,
    schema_module: ModuleType,
    schema_view: SchemaView,
) -> str:
    """Convert between different representations of schema:target_class instances

    The schema information is provided by `schema_module` and `schema_view`.
    Both can be created with `get_convertion_objects`
    """
    try:
        return _do_convert_format(
            target_class=target_class,
            data=data,
            input_format=input_format,
            output_format=output_format,
            schema_module=schema_module,
            schema_view=schema_view,
        )
    except Exception as e:  # BLE001
        msg = f'Conversion {input_format} -> {output_format} of data ({data}) failed for class {target_class}.'
        raise ValueError(msg) from e


def _do_convert_format(
    target_class: str,
    data: dict | str,
    input_format: Format,
    output_format: Format,
    schema_module: ModuleType,
    schema_view: SchemaView,
) -> str:
    """Convert between different representations of schema:target_class instances

    The schema information is provided by `schema_module` and `schema_view`.
    Both can be created with `get_convertion_objects`
    """

    if input_format == output_format:
        return data

    py_target_class = schema_module.__dict__[target_class]
    loader = get_loader(input_format.value)
    if input_format.value in ('ttl',):
        input_args = {'schemaview': schema_view, 'fmt': input_format.value}
    else:
        input_args = {}

    data_obj = loader.load(
        source=data,
        target_class=py_target_class,
        **input_args,
    )

    dumper = get_dumper(output_format.value)
    return dumper.dumps(
        data_obj, **({'schemaview': schema_view} if output_format == Format.ttl else {})
    )



from __future__ import annotations   # noqa: I001 -- the patches have to be imported early

import argparse
import logging
import sys
from pathlib import Path
from typing import (
    Annotated,  # noqa F401 -- used by generated code
    Any,
    TYPE_CHECKING,
)

from starlette.status import HTTP_413_REQUEST_ENTITY_TOO_LARGE

from dump_things_service.lazy_list import PriorityList, ModifierList
# Perform the patching before importing any third-party libraries
from dump_things_service.patches import enabled  # noqa: F401

import uvicorn
from fastapi import (
    Body,  # noqa F401 -- used by generated code
    Depends,
    FastAPI,
    HTTPException,
    Request,
)
from fastapi.middleware.cors import CORSMiddleware
from fastapi.security import APIKeyHeader
from fastapi_pagination import (
    Page,
    add_pagination,
    paginate,
)
from pydantic import (
    BaseModel,
    TypeAdapter,
)
from starlette.responses import (
    JSONResponse,
    PlainTextResponse,
)

from dump_things_service import (
    HTTP_400_BAD_REQUEST,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    Format,
    config_file_name,
)
from dump_things_service.config import (
    ConfigError,
    TokenPermission,
    InstanceConfig,
    get_default_token_name,
    get_token_store,
    get_zone,
    join_default_token_permissions,
    process_config,
)
from dump_things_service.converter import (
    FormatConverter,
    ConvertingList,
)
from dump_things_service.dynamic_endpoints import create_endpoints
from dump_things_service.export import exporter_info
from dump_things_service.model import (
    get_classes,
    get_subclasses,
)
from dump_things_service.utils import combine_ttl

if TYPE_CHECKING:
    from dump_things_service.lazy_list import LazyList
    from dump_things_service.store.model_store import ModelStore


class TokenCapabilityRequest(BaseModel):
    token: str | None


logger = logging.getLogger('dump_things_service')
uvicorn_logger = logging.getLogger('uvicorn')


parser = argparse.ArgumentParser()
parser.add_argument('--host', default='0.0.0.0')  # noqa S104
parser.add_argument('--port', default=8000, type=int)
parser.add_argument('--origins', action='append', default=[])
parser.add_argument(
    '-c', '--config',
    metavar='CONFIG_FILE',
    help="Read the configuration from 'CONFIG_FILE' instead of looking for it in the data store root directory. "
)
parser.add_argument(
    '--root-path',
    default='',
    help="Set the ASGI 'root_path' for applications submounted below a given URL path.",
)
parser.add_argument(
    '--error-mode',
    action='store_true',
    help="Don't exit with non-zero status on errors that prevent the service from proper operation, instead return the error on every request.",
)
parser.add_argument(
    '--log-level',
    default='WARNING',
    help="Set the log level for the service, allowed values are 'ERROR', 'WARNING', 'INFO', 'DEBUG'. Default is 'warning'.",
)
parser.add_argument(
    '--sort-by',
    action='append',
    default=[],
    help='Sort results by the given fields. Multiple fields can be specified, e.g. `--sort-by pid --sort-by date`.',
)
parser.add_argument(
    '--export-to',
    default='',
    metavar='FILE_NAME',
    help="Export the store to the file FILE_NAME and exit the process. If FILE_NAME is `-', the data is written to stdout.",
)
parser.add_argument(
    '--export-json',
    default='',
    metavar='FILE_NAME',
    help="Export the store to the file FILE_NAME and exit the process. If FILE_NAME is `-', the data is written to stdout.",
)
parser.add_argument(
    '--export-tree',
    default='',
    metavar='DIRECTORY_NAME',
    help="Export the store to a dumpthings-conformant tree at DIRECTORY_NAME and exit the process. If DIRECTORY_NAME does not exist, the service will try to create it.",
)
parser.add_argument(
    'store',
    help='The root of the data stores, it should contain a global_store and token_stores.',
)


arguments = parser.parse_args()

# Set the log level
numeric_level = getattr(logging, arguments.log_level.upper(), None)
if not isinstance(numeric_level, int):
    logger.error('Invalid log level: %s, defaulting to level "WARNING"', arguments.log_level)
else:
    logging.basicConfig(level=numeric_level)

store_path = Path(arguments.store)

g_error = None

config_path = Path(arguments.config) if arguments.config else store_path / config_file_name
try:
    g_instance_config = process_config(
        store_path=store_path,
        config_file=config_path,
        order_by=arguments.sort_by,
        globals_dict=globals(),
    )
except ConfigError:
    logger.exception(
        'ERROR: invalid configuration file at: `%s`',
        config_path,
    )
    g_error = 'Server runs in error mode due to an invalid configuration. See server error-log for details.'
    g_instance_config = None


for switch in ('to', 'json', 'tree'):
    argument = getattr(arguments, 'export_' + switch)
    if argument:
        if g_error:
            sys.stderr.write('ERROR: Configuration errors detected, cannot export store.')
            sys.exit(1)
        exporter_info[switch](g_instance_config, argument)
        sys.exit(0)


app = FastAPI()


def handle_global_error():
    if g_error:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=g_error)


# If a global error exists, it does not make sense to activate the defined
# endpoints because we don't have a working configuration. Instead, we signal
# the error to any request that is made to the server.
if g_error:
    if __name__ == '__main__' and arguments.error_mode:
        logger.warning('Server runs in error mode, all endpoints will return error information.')

        @app.post('/{full_path:path}')
        def post_global_error(request: Request, full_path: str):  # noqa: ARG001
            handle_global_error()

        @app.get('/{full_path:path}')
        def get_global_error(request: Request, full_path: str):  # noqa: ARG001
            handle_global_error()

        uvicorn.run(
            app,
            host=arguments.host,
            port=arguments.port,
            root_path=arguments.root_path,
        )
    sys.exit(1)


api_key_header_scheme = APIKeyHeader(
    name='X-DumpThings-Token',
    # authentication is generally optional
    auto_error=False,
    scheme_name='submission',
    description='Presenting a valid token enables record submission, and retrieval of records submitted with this token prior curation.',
)


def store_record(
    collection: str,
    data: BaseModel | str,
    class_name: str,
    model: Any,
    input_format: Format,
    api_key: str | None = Depends(api_key_header_scheme),
) -> JSONResponse | PlainTextResponse:
    if input_format == Format.json and isinstance(data, str):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail='Invalid JSON data provided.'
        )

    if input_format == Format.ttl and not isinstance(data, str):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail='Invalid ttl data provided.'
        )

    _check_collection(g_instance_config, collection)

    token = get_default_token_name(g_instance_config, collection) if api_key is None else api_key
    # Get the token permissions and extend them by the default permissions
    store, token_permissions = get_token_store(g_instance_config, collection, token)
    final_permissions = join_default_token_permissions(g_instance_config, token_permissions, collection)
    if not final_permissions.incoming_write:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail=f'Not authorized to submit to collection "{collection}".',
        )

    if input_format == Format.ttl:
        json_object = FormatConverter(
            g_instance_config.schemas[collection],
            input_format=Format.ttl,
            output_format=Format.json,
        ).convert(data, class_name)
        record = TypeAdapter(getattr(model, class_name)).validate_python(json_object)
    else:
        record = data

    stored_records = store.store_object(
        obj=record,
        submitter=g_instance_config.token_stores[token]['user_id'],
    )

    if input_format == Format.ttl:
        format_converter = FormatConverter(
            g_instance_config.schemas[collection],
            input_format=Format.json,
            output_format=Format.ttl,
        )
        return PlainTextResponse(
            combine_ttl(
                [
                    format_converter.convert(
                        record,
                        class_name,
                    )
                    for class_name, record in stored_records
                ]
            ),
            media_type='text/turtle',
        )
    return JSONResponse([record for _, record in stored_records])


def _check_collection(
    instance_config: InstanceConfig,
    collection: str,
):
    if collection not in instance_config.collections:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f'No such collection: "{collection}".',
        )


@app.post('/{collection}/token_permissions')
async def fetch_token_permissions(
    collection: str,
    body: TokenCapabilityRequest,
):
    _check_collection(g_instance_config, collection)
    token = get_default_token_name(g_instance_config, collection) if body.token is None else body.token
    token_store, token_permissions = get_token_store(g_instance_config, collection, token)
    final_permissions = join_default_token_permissions(g_instance_config, token_permissions, collection)
    return JSONResponse(
        {
            'read_curated': final_permissions.curated_read,
            'read_incoming': final_permissions.incoming_read,
            'write_incoming': final_permissions.incoming_write,
            **(
                {'incoming_zone': get_zone(g_instance_config, collection, token)}
                if final_permissions.incoming_read or final_permissions.incoming_write
                else {}
            ),
        }
    )


@app.get('/{collection}/record')
async def read_record_with_pid(
    collection: str,
    pid: str,
    format: Format = Format.json,  # noqa A002
    api_key: str = Depends(api_key_header_scheme),
):
    _check_collection(g_instance_config, collection)

    final_permissions, token_store = await process_token(g_instance_config, api_key, collection)

    class_name, json_object = None, None
    if final_permissions.incoming_read:
        class_name, json_object = token_store.get_object_by_pid(pid)

    if not json_object and final_permissions.curated_read:
        class_name, json_object = g_instance_config.curated_stores[collection].get_object_by_pid(pid)

    if not json_object:
        return None

    if format == Format.ttl:
        converter = FormatConverter(
            schema=g_instance_config.schemas[collection],
            input_format=Format.json,
            output_format=format,
        )
        ttl_record = converter.convert(json_object, class_name)
        return PlainTextResponse(ttl_record, media_type='text/turtle')
    return json_object


@app.get('/{collection}/records/{class_name}')
async def read_records_of_type(
    collection: str,
    class_name: str,
    format: Format = Format.json,  # noqa A002
    api_key: str = Depends(api_key_header_scheme),
):
    return await _read_records_of_type(
        collection=collection,
        class_name=class_name,
        format=format,
        api_key=api_key,
        # Set an upper limit for the number of non-paginated result records to
        # keep processing time for individual requests short and avoid
        # overloading the server.
        bound=1000,
    )


@app.get('/{collection}/records/p/{class_name}')
async def read_records_of_type_paginated(
        collection: str,
        class_name: str,
        format: Format = Format.json,  # noqa A002
        api_key: str = Depends(api_key_header_scheme),
) -> Page[dict | str]:

    result_list = await _read_records_of_type(
        collection=collection,
        class_name=class_name,
        format=format,
        api_key=api_key,
        bound=None,
    )
    return paginate(result_list)


async def _read_records_of_type(
        collection: str,
        class_name: str,
        format: Format = Format.json,  # noqa A002
        api_key: str = Depends(api_key_header_scheme),
        bound: int | None = None,
) -> LazyList:

    def check_bounds(length: int, max_length: int):
        if length > max_length:
            raise HTTPException(
                status_code=HTTP_413_REQUEST_ENTITY_TOO_LARGE,
                detail=f'Too many records found for class "{class_name}" in '
                       f'collection "{collection}". Please use pagination '
                       f'(/{collection}/records/p/{class_name}).',
            )

    _check_collection(g_instance_config, collection)

    model = g_instance_config.model_info[collection][0]
    if class_name not in get_classes(model):
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f'No "{class_name}"-class in collection "{collection}".',
        )

    final_permissions, token_store = await process_token(g_instance_config, api_key, collection)

    result_list = PriorityList()
    if final_permissions.incoming_read:
        for search_class_name in get_subclasses(model, class_name):
            token_store_list = token_store.get_objects_of_class(search_class_name)
            if bound:
                check_bounds(len(token_store_list), bound)
            result_list.add_list(token_store_list)

    if final_permissions.curated_read:
        for search_class_name in get_subclasses(model, class_name):
            curated_store_list = g_instance_config.curated_stores[collection].get_objects_of_class(search_class_name)
            if bound:
                check_bounds(len(curated_store_list), bound)
            result_list.add_list(curated_store_list)

    # Sort the result list.
    result_list.sort(key=result_list.sort_key)

    if format == Format.ttl:
        result_list = ConvertingList(
            result_list,
            g_instance_config.schemas[collection],
            input_format=Format.json,
            output_format=format,
        )
    else:
        result_list = ModifierList(
            result_list,
            lambda record_info: record_info.json_object,
        )
    return result_list


async def process_token(
    instance_config: InstanceConfig,
    api_key: str,
    collection: str,
) -> tuple[TokenPermission, ModelStore]:
    token = get_default_token_name(instance_config, collection) if api_key is None else api_key
    token_store, token_permissions = get_token_store(instance_config, collection, token)
    final_permissions = join_default_token_permissions(instance_config, token_permissions, collection)
    if not final_permissions.incoming_read and not final_permissions.curated_read:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail=f'No read access to curated or incoming data in collection "{collection}".',
        )
    return final_permissions, token_store


# If we have a valid configuration, create dynamic endpoints and rebuild the
# app to include all dynamically created endpoints.
if g_instance_config:
    create_endpoints(app, g_instance_config, globals())
    app.openapi_schema = None
    app.setup()

# Add CORS origins
app.add_middleware(
    CORSMiddleware,
    allow_origins=arguments.origins,
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)

# Add pagination
add_pagination(app)


def main():
    uvicorn.run(
        app,
        host=arguments.host,
        port=arguments.port,
        root_path=arguments.root_path,
    )


if __name__ == '__main__':
    main()



from __future__ import annotations

import dataclasses  # noqa F401 -- used by generated code
import importlib
import logging
import subprocess
import tempfile
from itertools import count
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
)

import annotated_types  # noqa F401 -- used by generated code
import pydantic  # noqa F401 -- used by generated code
import pydantic_core  # noqa F401 -- used by generated code
from linkml.generators import PythonGenerator
from linkml_runtime import SchemaView
from pydantic._internal._model_construction import ModelMetaclass

from dump_things_service.utils import (
    read_url,
    sys_path,
)

if TYPE_CHECKING:
    from types import ModuleType


lgr = logging.getLogger('uvicorn')

serial_number = count()
_model_counter = count()

_model_cache = {}
_schema_model_cache = {}
_schema_view_cache = {}


def build_model(
    source_url: str,
) -> Any:
    with tempfile.TemporaryDirectory() as temp_dir:
        module_name = f'model_{next(serial_number)}'
        definition_file = Path(temp_dir) / 'definition.yaml'
        definition_file.write_text(read_url(source_url))
        subprocess.run(
            args=['gen-pydantic', str(definition_file)],
            stdout=(Path(temp_dir) / (module_name + '.py')).open('w'),
            check=True,
        )
        with sys_path([temp_dir]):
            model = importlib.import_module(module_name)
    return model


def get_classes(
    model: Any,
) -> list[str]:
    """get names of all subclasses of Thing"""
    return get_subclasses(model, 'Thing')


def get_subclasses(
    model: ModuleType,
    class_name: str,
) -> list[str]:
    """get names of all subclasses (includes class_name itself)"""
    super_class = getattr(model, class_name)
    return [
        name
        for name, obj in model.__dict__.items()
        if isinstance(obj, ModelMetaclass) and issubclass(obj, super_class)
    ]


def get_model_for_schema(
    schema_location: str,
) -> tuple[ModuleType, list[str], str]:
    if schema_location not in _model_cache:
        lgr.info(f'Building model for schema {schema_location}.')
        model = build_model(schema_location)
        classes = get_classes(model)
        model_var_name = f'model_{next(_model_counter)}'
        _model_cache[schema_location] = model, classes, model_var_name
    return _model_cache[schema_location]


def get_schema_view(schema_location: str) -> SchemaView:
    if schema_location not in _schema_view_cache:
        _schema_view_cache[schema_location] = SchemaView(schema_location)
    return _schema_view_cache[schema_location]


def get_schema_model_for_schema(
    schema_location: str,
) -> ModuleType:
    if schema_location not in _schema_model_cache:
        _schema_model_cache[schema_location] = PythonGenerator(schema_location).compile_module()
    return _schema_model_cache[schema_location]



""" Implementation of a lazy list

The lazy list calls a subclass method to generate list elements on demand. The
generation is based on the list index and information stored by the user of the
list via `add_info`. The list has as many entries as information entries were
added via `add_info`.

Lazy lists can be used to refer to large number of records with non-trivial
storage and retrieval costs. It is mainly used to:

 1. implement priorities between results of different stores,
 2. implement result-sorting across different stores,
 3. efficiently implement result pagination with the simple `paginate`-interface
    of `fastapi-pagination`, which requires a list-like object that holds all
    results.

 The memory footprint of a lazy list depends on the number of entries and the
 size of the user-supplied `info`-object.
"""
from __future__ import annotations

from abc import (
    ABCMeta,
    abstractmethod,
)
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import (
        Any,
        Callable,
    )


class LazyList(list, metaclass=ABCMeta):

    class LazyListIterator:
        def __init__(self, lazy_list: LazyList):
            self.lazy_list = lazy_list
            self.index = 0

        def __iter__(self) -> LazyList.LazyListIterator:
            return self

        def __next__(self) -> str | dict:
            if self.index == len(self.lazy_list):
                raise StopIteration
            self.index += 1
            return self.lazy_list[self.index - 1]

    def __init__(self):
        super().__init__()
        self.list_info = []

    def __iter__(self) -> LazyList.LazyListIterator:
        return LazyList.LazyListIterator(self)

    def __len__(self) -> int:
        return len(self.list_info)

    def __getitem__(self, index: int) -> Any:
        if isinstance(index, slice):
            start = 0 if index.start is None else index.start
            stop = len(self) if index.stop is None else index.stop
            step = 1 if index.step is None else index.step

            # There is a condition in the list comprehension that ensures that
            # the indices are within bounds. We still adjust them here to
            # reduce the number of iterations in the list comprehension.
            if start < 0:
                start = max(len(self) + start, 0)
            elif start >= len(self):
                start = len(self) - 1

            if stop < 0:
                stop = max(len(self) + stop, -1)
            elif stop > len(self):
                stop = len(self)

            return [
                self.generate_element(i, self.list_info[i])
                for i in range(start, stop, step)
                if 0 <= i < len(self.list_info)
            ]

        return self.generate_element(index, self.list_info[index])

    @abstractmethod
    def generate_element(self, index: int, info: Any) -> Any:
        """
        Generate the list element at the specified index, using stored `info`

        This method should be implemented by subclasses to retrieve the content
        of the record at the given index. The generation is usually based on the
        `index` and the `info` object stored that was previously stored via
        `add_info`.

        :param index: The index of the record to retrieve.
        :param info: The object stored at the specified index in the info list.
        :return: The full element at the specified index, usually generated using
            `index` and `info`.
        """
        raise NotImplementedError

    def unique_identifier(self, info: Any) -> Any:
        """
        Return a unique identifier for the represented information

        The unique identifier is used in priority lists to uniquely identify an
        object across multiple lists.

        This should be implemented if the list is supposed to be used in a
        priority list.

        :param info: The surrogate information.
        :return: A unique identifier for the element represented by the surrogate.
        """
        raise NotImplementedError

    def sort_key(self, info: Any) -> str:
        """
        Return a string that can be used for sorting the list.

        This should be implemented if the list is supposed to be sorted.

        :param info: The surrogate information.
        :return: The string that should be used for sorting the list.
        """
        raise NotImplementedError

    def add_info(
        self,
        info: Iterable[Any],
    ) -> LazyList:
        """
        Add list entry information to the list.

        :param info: An iterable that contains opaque information that will be
            given to `generate_element` in the `info`-argument and can be used
            to generate a list element on demand. `info` could, for example, be
            a database-ID or a file path.
        """
        self.list_info.extend(info)
        return self

    def sort(
        self,
        *,
        key: Callable | None = None,
        reverse: bool = False,
    ) -> LazyList:
        """
        Sort the lazy list based on a key function.
        """
        self.list_info.sort(key=key, reverse=reverse)
        return self

    def get_key_function(self) -> Callable:
        """
        Get the key function used to sort the list.

        This method should be implemented by subclasses to return a function
        that can be used to extract a key from the list info for sorting.
        """
        raise NotImplementedError


class PriorityList(LazyList):
    """
    A lazy list that emits every item, identified by `key` only once.

    Items from lists the were added earlier have a higher priority than items
    from lists that were added later. Items are identified via the
    `unique_identifier` method of the added lists.
    All lists should be added before the first iteration.
    """
    def __init__(
        self,
    ):
        super().__init__()
        self.seen = set()
        self.type = None

    def add_list(
        self,
        input_list: LazyList,
    ) -> PriorityList:
        # Check the type
        if self.type:
            if not isinstance(input_list, self.type):
                msg = f'Expected input_list of type {self.type}, got {type(input_list)}'
                raise TypeError(msg)
        else:
            self.type = type(input_list)

        for info in input_list.list_info:
            criteria = input_list.unique_identifier(info)
            if criteria not in self.seen:
                self.seen.add(criteria)
                self.list_info.append((info, input_list))
        return self

    def generate_element(self, index: int, info: Any) -> Any:
        # Delegate the generation to the input list
        return info[1].generate_element(index, info[0])

    def sort_key(self, info: Any) -> str:
        # Delegate the sort key to the input list
        return info[1].sort_key(info[0])


class ModifierList(LazyList):
    """
    A lazy list that modifies every item of `input_list` by the `modifier`
    """
    def __init__(
            self,
            input_list: LazyList,
            modifier: Callable,
    ):
        super().__init__()
        self.input_list = input_list
        self.modifier = modifier
        self.list_info = input_list.list_info

    def generate_element(self, index: int, info: Any) -> Any:
        return self.modifier(self.input_list.generate_element(index, info))



"""Monkeypatch for linkml_runtime.loaders.rdflib_loader.RDFLibLoader.from_rdf_graph

Corresponds to the patch `patches/rdflib_loader_typedesignator.diff` in the
`datalad-concepts` repository, i.e.,
<https://github.com/psychoinformatics-de/datalad-concepts.git>
"""

from __future__ import annotations

import logging
from copy import copy
from dataclasses import dataclass
from importlib import import_module
from typing import (
    Any,
    Union,
)

from curies import Converter
from linkml_runtime import MappingError
from linkml_runtime.linkml_model import (
    ClassDefinition,
    ClassDefinitionName,
    EnumDefinition,
    SlotDefinition,
    TypeDefinition,
)
from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.utils.schemaview import SchemaView
from linkml_runtime.utils.yamlutils import YAMLRoot
from pydantic import BaseModel
from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import RDF
from rdflib.term import (
    BNode,
    Literal,
)

logger = logging.getLogger(__name__)


VALID_SUBJECT = Union[URIRef, BNode]
ANYDICT = dict[str, Any]


@dataclass
class Pointer:
    obj: str


def from_rdf_graph(
        self,
        graph: Graph,
        schemaview: SchemaView,
        target_class: type[Union[BaseModel, YAMLRoot]],
        prefix_map: Union[dict[str, str], Converter, None] = None,
        cast_literals: bool = True,
        allow_unprocessed_triples: bool = True,
        ignore_unmapped_predicates: bool = False,
) -> list[Union[BaseModel, YAMLRoot]]:
    """
    Loads objects from graph into lists of the python target_class structure,
    recursively walking RDF graph from instances of target_class.

    :param graph: rdflib Graph that holds instances of target_class
    :param schemaview: schema to which graph conforms
    :param target_class: class which root nodes should instantiate
    :param prefix_map: additional prefix mappings for data objects
    :param ignore_unmapped_predicates: if True then a predicate that has no mapping to a slot does not raise an error
    :return: all instances of target class type
    """
    namespaces = schemaview.namespaces()
    uri_to_class_map = {}
    for cn, c in schemaview.all_classes().items():
        uri = schemaview.get_uri(c, expand=True)
        if uri in uri_to_class_map:
            c2 = uri_to_class_map[uri]
            if c2.name in schemaview.class_ancestors(cn):
                continue
            else:
                logger.error(f'Inconsistent URI to class map: {uri} -> {c2.name}, {c.name}')
        uri_to_class_map[uri] = c
    # data prefix map: supplements or overrides existing schema prefix map
    if isinstance(prefix_map, Converter):
        # TODO replace with `prefix_map = prefix_map.bimap` after making minimum requirement on python 3.8
        prefix_map = {record.prefix: record.uri_prefix for record in prefix_map.records}
    if prefix_map:
        for k, v in prefix_map.items():
            namespaces[k] = v
            graph.namespace_manager.bind(k, URIRef(v))
    # Step 1: Create stub root dict-objects
    target_class_uriref: URIRef = target_class.class_class_uri
    root_dicts: list[ANYDICT] = []
    root_subjects: list[VALID_SUBJECT] = list(graph.subjects(RDF.type, target_class_uriref))
    logger.debug(f'ROOTS = {root_subjects}')
    # Step 2: walk RDF graph starting from root subjects, constructing dict tree
    node_tuples_to_visit: list[tuple[VALID_SUBJECT, ClassDefinitionName]]  ## nodes and their type still to visit
    node_tuples_to_visit = [(subject, target_class.class_name) for subject in root_subjects]
    uri_to_slot: dict[str, SlotDefinition]  ## lookup table for RDF predicates -> slots
    uri_to_slot = {URIRef(schemaview.get_uri(s, expand=True)): s for s in schemaview.all_slots().values()}
    processed: set[VALID_SUBJECT] = set()  ## track nodes already visited, or already scheduled
    for n, _ in node_tuples_to_visit:
        processed.add(n)
    obj_map: dict[VALID_SUBJECT, ANYDICT] = {}  ## map from an RDF node to its dict representation
    unmapped_predicates = set()
    processed_triples: set[tuple] = set()
    while len(node_tuples_to_visit) > 0:
        subject, subject_class = node_tuples_to_visit.pop()
        processed.add(subject)
        dict_obj = self._get_id_dict(subject, schemaview, subject_class)
        if subject in root_subjects:
            root_dicts.append(dict_obj)
        obj_map[subject] = dict_obj
        type_designator_slot = schemaview.get_type_designator_slot(subject_class)
        if type_designator_slot:
            td_iri = schemaview.get_uri(type_designator_slot, expand=True)
            type_vals = list(graph.objects(subject, URIRef(td_iri)))
            if len(type_vals) > 0:
                type_classes = [uri_to_class_map[str(x)] for x in type_vals]
                if len(type_classes) > 1:
                    raise ValueError(f'Ambiguous types for {subject} == {type_classes}')
                logger.info(f'Replacing {subject_class} with {type_classes}')
                subject_class = type_classes[0].name
                # PATCH >>>>>
                if type_classes[0].class_uri is not None:
                    dict_obj[type_designator_slot.name] = type_classes[0].class_uri
                # PATCH <<<<<
        # process all triples for this node
        for (_, p, o) in graph.triples((subject, None, None)):
            processed_triples.add((subject,p,o))
            logger.debug(f' Processing triple {subject} {p} {o}, subject type = {subject_class}')
            if p == RDF.type:
                logger.debug(f'Ignoring RDF.type for {subject} {o}, we automatically infer this from {subject_class}')
            elif p not in uri_to_slot:
                if ignore_unmapped_predicates:
                    unmapped_predicates.add(p)
                else:
                    raise MappingError(f'No pred for {p} {type(p)}')
            else:
                slot = schemaview.induced_slot(uri_to_slot[p].name, subject_class)
                range_applicable_elements = schemaview.slot_applicable_range_elements(slot)
                is_inlined = schemaview.is_inlined(slot)
                slot_name = underscore(slot.name)
                if isinstance(o, Literal):
                    if EnumDefinition.class_name in range_applicable_elements:
                        logger.debug(f'Assuming no meaning assigned for value {o} for Enum {slot.range}')
                    elif TypeDefinition.class_name not in range_applicable_elements:
                        raise ValueError(f'Cannot map Literal {o} to a slot {slot.name} whose range {slot.range} is not a type;')
                    v = o.value
                elif isinstance(o, BNode):
                    if not is_inlined:
                        logger.error(f'blank nodes should be inlined; {slot_name}={o} in {subject}')
                    v = Pointer(o)
                else:
                    if ClassDefinition.class_name in range_applicable_elements:
                        if slot.range in schemaview.all_classes():
                            id_slot = schemaview.get_identifier_slot(slot.range)
                            v = self._uri_to_id(o, id_slot, schemaview)
                        else:
                            v = namespaces.curie_for(o)
                        if v is None:
                            logger.debug(f'No CURIE for {p}={o} in {subject} [{subject_class}]')
                            v = str(o)
                    elif EnumDefinition.class_name in range_applicable_elements:
                        range_union_elements = schemaview.slot_range_as_union(slot)
                        enum_names = [e for e in range_union_elements if e in schemaview.all_enums()]
                        # if a PV has a meaning URI declared, map this
                        # back to a text representation
                        v = namespaces.curie_for(o)
                        for enum_name in enum_names:
                            e = schemaview.get_enum(enum_name)
                            for pv in e.permissible_values.values():
                                if v == pv.meaning or str(o) == pv.meaning:
                                    v = pv.text
                                    break
                    elif TypeDefinition.class_name in range_applicable_elements:
                        if cast_literals:
                            v = namespaces.curie_for(o)
                            if v is None:
                                v = str(o)
                            logger.debug(f'Casting {o} to string')
                        else:
                            raise ValueError(f'Expected literal value ({range_applicable_elements}) for {slot_name}={o}')
                    if is_inlined:
                        # the object of the triple may not yet be processed;
                        # we store a pointer to o, and then replace this later
                        v = Pointer(o)
                if slot.multivalued:
                    if slot_name not in dict_obj:
                        dict_obj[slot_name] = []
                    dict_obj[slot_name].append(v)
                else:
                    dict_obj[slot_name] = v
                if o not in processed:
                    # if o instantiates a class, add to list of nodes to be visited.
                    # force type based on range constraint
                    if slot.range in schemaview.all_classes():
                        node_tuples_to_visit.append((o, ClassDefinitionName(slot.range)))
    if unmapped_predicates:
        logger.info(f'Unmapped predicated: {unmapped_predicates}')
    unprocessed_triples = set(graph.triples((None, None, None))) - processed_triples
    logger.info(f'Triple processed = {len(processed_triples)}, unprocessed = {len(unprocessed_triples)}')
    if len(unprocessed_triples) > 0:
        if not allow_unprocessed_triples:
            for t in unprocessed_triples:
                logger.warning(f'  Unprocessed: {t}')
            raise ValueError(f'Unprocessed triples: {len(unprocessed_triples)}')
    # Step 2: replace inline pointers with object dicts
    def repl(v):
        if isinstance(v, Pointer):
            v2 = obj_map[v.obj]
            if v2 is None:
                raise Exception(f'No mapping for pointer {v}')
            return v2
        else:
            return v
    objs_to_visit: list[ANYDICT] = copy(root_dicts)
    while len(objs_to_visit) > 0:
        obj = objs_to_visit.pop()
        logger.debug(f'Replacing pointers for  {obj}')
        for k, v in obj.items():
            if v is None:
                continue
            if isinstance(v, list):
                v = [repl(v1) for v1 in v if v1 is not None]
                for v1 in v:
                    if isinstance(v1, dict):
                        objs_to_visit.append(v1)
            else:
                v = repl(v)
                if isinstance(v, dict):
                    objs_to_visit.append(v)
            obj[k] = v
    # Final step: translate dicts into instances of target_class
    return [target_class(**x) for x in root_dicts]


mod = import_module('linkml_runtime.loaders.rdflib_loader')
mod.RDFLibLoader.from_rdf_graph = from_rdf_graph



"""
Base class for storage backends

Storage backends return multiple records as `LazyList[RecordInfo]` objects.
The reason for using a lazy list instead of yielding records one by one is that
fastapi endpoints and fastapi-pagination work with list like objects and not
with generators, i.e. it uses index- or slice-based access to the records.
"""

from __future__ import annotations

from abc import (
    ABCMeta,
    abstractmethod,
)
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
)

from dump_things_service.lazy_list import LazyList

if TYPE_CHECKING:
    from collections.abc import Iterable


@dataclass
class RecordInfo:
    iri: str
    class_name: str
    json_object: dict[str, Any]
    # We store a sort key to support sorting of records from multiple, probably
    # sorted, sources.
    sort_key: str


@dataclass
class ResultListInfo:
    iri: str
    class_name: str
    sort_key: str
    private: Any


class BackendResultList(LazyList):
    """
    Implementation of a lazy list that holds references to records stored in
    a backend. The list elements carry a sort key that allows sorting of the
    list. The lazy list-approach is used for the results of
    `get_records_of_classes`, which can be large, i.e. in the millions, because:

    1. Fastapi pagination requires a list-like object that can be sliced and
       indexed. Lazy lists allow us to keep only objects of the current slice
       in memory.

    2. The service integrates elements from multiple backends; this requires
       a possibility to sort the results from different backends into an
       integrated result. The `sort_key` supports this by providing a
       backend-independent, record-specific key that can be used to sort the
       records.
        """

    def generate_element(self, index: int, info: ResultListInfo) -> RecordInfo:
        """
        Generate a JSON representation of the record at index `index`.

        :param index: The index of the record that should be retrieved (ignored).
        :param info: The tuple (iri, record_class_name, record_path).
        :return: A JSON object.
        """
        return self.generate_result(
            index,
            info.iri,
            info.class_name,
            info.sort_key,
            info.private
        )

    def unique_identifier(self, info: ResultListInfo) -> Any:
        # Return the IRI as unique identifier
        return info.iri

    def sort_key(self, info: ResultListInfo) -> str:
        # Return the sort_key entry as sort key
        return info.sort_key

    @abstractmethod
    def generate_result(
        self,
        index: int,
        iri: str,
        class_name: str,
        sort_key: str,
        private: Any,
    ) -> RecordInfo:
        """
        Generate a record info object from the provided parameters.

        :param index: The index of the record.
        :param iri: The IRI of the record.
        :param class_name: The class name of the record.
        :param sort_key: The sort key for the record.
        :param private: Additional private information, if any.
        :return: A RecordInfo object.
        """
        raise NotImplementedError


class StorageBackend(metaclass=ABCMeta):
    def __init__(
        self,
        order_by: Iterable[str] | None = None,
    ):
        self.order_by = order_by or ['pid']

    @abstractmethod
    def add_record(
        self,
        iri: str,
        class_name: str,
        json_object: dict,
    ):
        raise NotImplementedError

    def add_records_bulk(
        self,
        object_info: Iterable[RecordInfo],
    ):
        """Default implementation for adding multiple records at once."""
        for info in object_info:
            self.add_record(
                iri=info.iri,
                class_name=info.class_name,
                json_object=info.json_object,
            )

    @abstractmethod
    def get_record_by_iri(
        self,
        iri: str,
    ) -> RecordInfo | None:
        raise NotImplementedError

    @abstractmethod
    def get_records_of_classes(
        self,
        class_names: Iterable[str],
    ) -> LazyList[RecordInfo]:
        raise NotImplementedError


def create_sort_key(
    json_object: dict[str, Any],
    order_by: Iterable[str],
) -> str:
    return '-'.join(
        str(json_object.get(key))
        if json_object.get(key) is not None
        else chr(0x10FFFF)
        for key in order_by
    )



from pathlib import Path

import yaml

from dump_things_service.config import (
    InstanceConfig,
    get_mapping_function_by_name,
)
from dump_things_service.model import get_classes
from dump_things_service.store.model_store import ModelStore


idfx = get_mapping_function_by_name('digest-md5-p3-p3')


def export_tree(
    instance_config: InstanceConfig,
    destination: str,
):
    destination = Path(destination)
    if destination.exists() and not destination.is_dir():
        msg = 'The export_tree destination path must be a directory.'
        raise ValueError(msg)

    destination.mkdir(parents=True, exist_ok=True)
    for collection in instance_config.collections:
        export_collection(
            instance_config,
            collection,
            destination,
        )


def export_collection(
    instance_config: InstanceConfig,
    collection: str,
    destination: Path,
):
    collection_destination = destination / collection
    collection_destination.mkdir(parents=True, exist_ok=True)

    config_content = (
        "type: records\n"
        "version: 1\n"
        f"schema: {instance_config.schemas[collection]}\n"
        "format: yaml\n"
        "idfx: digest-md5-p3-p3\n"
    )

    curated_destination = collection_destination / 'curated'
    curated_destination.mkdir(parents=True, exist_ok=True)
    (curated_destination / '.dumpthings.yaml').write_text(config_content)
    exported_stores = {
        id(instance_config.curated_stores[collection]): curated_destination
    }
    export_classes(instance_config.curated_stores[collection], curated_destination)

    # Determine stores for incoming zones
    zones = {
        label: instance_config.token_stores[token]['collections'].get(collection, {}).get('store')
        for token, label in instance_config.zones.get(collection, {}).items()
        if instance_config.token_stores[token]['collections'].get(collection, {}).get('store') is not None
    }

    if zones:
        incoming_destination = collection_destination / 'incoming'
        for zone, store in zones.items():
            zone_destination = incoming_destination / zone
            if id(store) in exported_stores:
                # Already exported this store, make `zone_destination` a link
                # to the existing export.
                zone_destination.parent.mkdir(parents=True, exist_ok=True)
                zone_destination.symlink_to(exported_stores[id(store)])
                continue
            exported_stores[id(store)] = zone_destination = collection_destination / 'incoming' / zone
            zone_destination.mkdir(parents=True, exist_ok=True)
            (zone_destination / '.dumpthings.yaml').write_text(config_content)
            export_classes(store, zone_destination)


def export_classes(
    store: ModelStore,
    destination: Path,
):
    class_names = get_classes(store.model)
    for class_name in class_names:

        # We know that pure `Thing` instances are not stored in the store.
        if class_name == 'Thing':
            continue

        record_infos = store.get_objects_of_class(class_name, include_subclasses=False)
        if record_infos:
            class_destination = destination / class_name
            class_destination.mkdir(parents=True, exist_ok=True)
            for record_info in record_infos:
                json_object = record_info.json_object
                instance_destination = class_destination / idfx(
                    json_object['pid'],
                    'yaml',
                )
                instance_destination.parent.mkdir(parents=True, exist_ok=True)
                instance_destination.write_text(
                    yaml.dump(
                        data=json_object,
                        sort_keys=False,
                        allow_unicode=True,
                        default_flow_style=False,
                    )
                )



from dump_things_service.tests.fixtures import (
    dump_stores_simple,
    fastapi_app_simple,
    fastapi_client_simple,
)

__all__ = [
    'dump_stores_simple',
    'fastapi_app_simple',
    'fastapi_client_simple',
]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import types


def resolve_curie(
    model: types.ModuleType,
    curie: str,
) -> str:
    if ":" not in curie:
        return curie

    if curie.startswith(('http://', 'https://')):
        return curie

    prefix, identifier = curie.split(':', 1)
    prefix_value = model.linkml_meta.root.get('prefixes', {}).get(prefix)
    if prefix_value is None:
        return curie
    return prefix_value['prefix_reference'] + identifier



import json
import sys
from pathlib import Path
from typing import TextIO

from dump_things_service.config import InstanceConfig
from dump_things_service.lazy_list import LazyList
from dump_things_service.model import get_classes
from dump_things_service.store.model_store import ModelStore

level_width = 2


# The _lookahead function is taken from:
# https://stackoverflow.com/questions/1630320/what-is-the-pythonic-way-to-detect-the-last-element-in-a-for-loop
# with small changes
def _lookahead(iterable):
    """Pass through all values from the given iterable, augmented by the
    information if there are more values to come after the current one
    (True), or if it is the last value (False).
    """
    # Get an iterator and pull the first value.
    it = iter(iterable)
    try:
        last = next(it)
    except StopIteration:
        return
    # Run the iterator to exhaustion (starting from the second value).
    for val in it:
        # Report the *previous* value (more to come).
        yield last, False
        last = val
    # Report the last value.
    yield last, True


def export_json(
    instance_config: InstanceConfig,
    destination: str,
):
    if destination == '-':
        output = sys.stdout
    else:
        output = Path(destination).open('wt', encoding='utf-8')

    output.write('{\n')
    for collection, is_last in _lookahead(instance_config.collections):
        output.write(f'{level_width * " "}"{collection}": {{\n')
        export_collection(instance_config, collection, 2 * level_width, output)
        if is_last:
            output.write(f'\n{level_width * " "}}}\n')
        else:
            output.write(f'\n{level_width * " "}}},\n')
    output.write('}\n')


def export_collection(
    instance_config: InstanceConfig,
    collection: str,
    indent: int,
    output: TextIO,
):
    output.write(f'{indent * " "}"schema": "{instance_config.schemas[collection]}",\n')
    output.write(f'{indent * " "}"curated": {{\n')
    append_classes(instance_config.curated_stores[collection], indent + level_width, output)
    output.write(f'\n{indent * " "}}}')

    # Determine stores for incoming zones
    zones = {
        label: instance_config.token_stores[token]['collections'].get(collection, {}).get('store')
        for token, label in instance_config.zones.get(collection, {}).items()
        if instance_config.token_stores[token]['collections'].get(collection, {}).get('store') is not None
    }

    if zones:
        # Put a comma between "curated" and "incoming".
        output.write(f',\n{indent * " "}"incoming": {{\n')
        indent_zone = indent + level_width
        indent_classes = indent_zone + level_width
        for (zone, store), is_last in _lookahead(zones.items()):
            output.write(f'{indent_zone * " "}"{zone}": {{\n')
            append_classes(store, indent_classes, output)
            if is_last:
                output.write(f'\n{(indent + level_width) * " "}}}')
            else:
                output.write(f'\n{(indent + level_width) * " "}}},\n')

        # End the "incoming" dictionary
        output.write(f'\n{indent * " "}}}')


def append_classes(
    store: ModelStore,
    indent: int,
    output: TextIO,
):
    """Append instances of all classes to the file"""
    class_names = get_classes(store.model)

    first = True
    for class_name in class_names:

        # We know that pure `Thing` instances are not stored in the store.
        if class_name == 'Thing':
            continue

        class_instances = store.get_objects_of_class(class_name, include_subclasses=False)
        if class_instances:
            if not first:
                output.write(',\n')
            first = False
            output.write(f'{indent * " "}"{class_name}": [\n')
            append_instances(
                class_instances,
                output,
                indent + level_width,
            )
            output.write(f'\n{indent * " "}]')


def append_instances(
    instances: LazyList,
    output: TextIO,
    indent: int,
):
    for instance, is_last in _lookahead(instances):
        json_string = json.dumps(instance.json_object, ensure_ascii=False)
        output.write(f'{(indent + level_width) * " "}{json_string}')
        if not is_last:
            output.write(',\n')



import logging
from itertools import count

from fastapi import FastAPI

from dump_things_service.config import InstanceConfig

logger = logging.getLogger('dump_things_service')


_endpoint_template = """
async def {name}(
        data: {model_var_name}.{class_name} | Annotated[str, Body(media_type='text/plain')],
        api_key: str = Depends(api_key_header_scheme),
        format: Format = Format.json,
) -> JSONResponse | PlainTextResponse:
    uvicorn_logger.info('{name}(%s, %s, %s, %s)', repr(data), repr('{class_name}'), repr({model_var_name}), repr(format))
    return store_record('{collection}', data, '{class_name}', {model_var_name}, format, api_key)
"""


def create_endpoints(
    app: FastAPI,
    instance_config: InstanceConfig,
    global_dict: dict,
):
    # Create endpoints for all classes in all collections
    logger.info('Creating dynamic endpoints...')
    serial_number = count()

    for collection, (model, classes, model_var_name) in instance_config.model_info.items():
        global_dict[model_var_name] = model
        for class_name in classes:
            # Create an endpoint to dump data of type `class_name` in version
            # `version` of schema `application`.
            endpoint_name = f'_endpoint_{next(serial_number)}'

            endpoint_source = _endpoint_template.format(
                name=endpoint_name,
                model_var_name=model_var_name,
                class_name=class_name,
                collection=collection,
                info=f"'store {collection}/{class_name} objects'",
            )
            exec(endpoint_source, global_dict)  # noqa S102

            # Create an API route for the endpoint
            app.add_api_route(
                path=f'/{collection}/record/{class_name}',
                endpoint=global_dict[endpoint_name],
                methods=['POST'],
                name=f'handle "{class_name}" of schema "{model.linkml_meta["id"]}" objects',
                response_model=None,
            )

    logger.info('Creation of %d endpoints completed.', next(serial_number))



from __future__ import annotations

import sys
from contextlib import contextmanager
from functools import reduce
from typing import TYPE_CHECKING

import fsspec
from rdflib import Graph

if TYPE_CHECKING:
    from pathlib import Path

    from dump_things_service import JSON
    from dump_things_service.config import InstanceConfig


@contextmanager
def sys_path(paths: list[str | Path]):
    """Patch the `Path` class to return the paths in `paths` in order."""
    original_path = sys.path
    try:
        sys.path = [str(path) for path in paths]
        yield
    finally:
        sys.path = original_path


def read_url(url: str) -> str:
    """
    Read the content of an URL into memory.
    """
    open_file = fsspec.open(url, 'rt')
    with open_file as f:
        return f.read()


def cleaned_json(
    data: JSON,
    remove_keys: tuple[str, ...] = ('@type',)
) -> JSON:
    if isinstance(data, list):
        return [cleaned_json(item, remove_keys) for item in data]
    if isinstance(data, dict):
        return {
            key: cleaned_json(value, remove_keys)
            for key, value in data.items()
            if key not in remove_keys and data[key] is not None
        }
    return data


def combine_ttl(documents: list[str]) -> str:
    graphs = [Graph().parse(data=doc, format='ttl') for doc in documents]
    return reduce(lambda g1, g2: g1 + g2, graphs).serialize(format='ttl')


def get_schema_type_curie(
    instance_config: InstanceConfig,
    collection: str,
    class_name: str,
) -> str:
    schema_url = instance_config.schemas[collection]
    schema_module = instance_config.conversion_objects[schema_url]['schema_module']
    class_object = getattr(schema_module, class_name)
    return class_object.class_class_curie
